import pytest
import torch

from higgs_kernels.hadamard import hadamard_transform


def test_hadamard_matrix_4():
    reference = (
        torch.tensor([1, 1, 1, 1, 1, -1, 1, -1, 1, 1, -1, -1, 1, -1, -1, 1])
        .reshape(4, 4)
        .to(torch.int64)
    )
    torch.testing.assert_close(
        hadamard_transform(torch.eye(4, dtype=torch.int64)), reference
    )


def hadamard_transform_matmul(x):
    out_dim, in_dim = x.shape
    return x @ hadamard_transform(torch.eye(in_dim, dtype=x.dtype))


@pytest.mark.parametrize("in_dim", [1, 2, 4, 128])
@pytest.mark.parametrize("out_dim", [1, 2, 3])
def test_hadamard(in_dim, out_dim):
    x = torch.randn((out_dim, in_dim), dtype=torch.float32)
    torch.testing.assert_close(hadamard_transform(x), hadamard_transform_matmul(x))



import pytest
import torch

from higgs_kernels import higgs_dequantize_2_256_kernel, higgs_quantize_2_256_kernel
from higgs_kernels.references import (
    higgs_dequantize_2_256_torch,
    higgs_quantize_2_256_torch,
)


@pytest.mark.parametrize("N", [1, 2, 3, 4, 8, 500, 512, 1023, 1024, 1025, 1024**2])
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
def test_dequantize(N, dtype):
    grid = torch.randn((256, 2), device="cuda", dtype=dtype)

    quantized = torch.randint(0, 256, (N,), dtype=torch.uint8, device="cuda")

    torch.testing.assert_close(
        higgs_dequantize_2_256_torch(quantized, grid),
        higgs_dequantize_2_256_kernel(quantized, grid),
    )


@pytest.mark.parametrize("N", [1, 2, 3, 4, 8, 500, 512, 1023, 1024, 1025, 1024**2])
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
def test_quantize(N, dtype):
    grid = torch.randn((256, 2), device="cuda", dtype=dtype)
    grid_norms = torch.linalg.norm(grid, dim=1) ** 2

    weight = torch.randn((N, 2), device="cuda", dtype=dtype)

    torch.testing.assert_close(
        higgs_quantize_2_256_torch(weight, grid, grid_norms),
        higgs_quantize_2_256_kernel(weight, grid, grid_norms),
    )



import torch

from higgs_kernels.grids import load_optimal_grid_2_256
from higgs_kernels.linear import (
    higgs_dequantize_linear_2_256,
    higgs_quantize_linear_2_256,
)


def higgs_quantize_dequantize(weight, grid, group_size=512):
    quantized, scales = higgs_quantize_linear_2_256(weight, grid, group_size)
    return higgs_dequantize_linear_2_256(quantized, scales, grid, group_size)


def test_quantize_linear():
    weight = torch.randn((4096, 1024), dtype=torch.float16, device="cuda")
    grid = load_optimal_grid_2_256(dtype=torch.float16, device="cuda")

    new_weight = higgs_quantize_dequantize(weight, grid)
    assert torch.linalg.norm(new_weight - weight) / torch.linalg.norm(weight) < 0.15



import torch


def is_pow2(n: int) -> bool:
    return n > 0 and (n & (n - 1)) == 0


@torch.no_grad()
def hadamard_transform(x: torch.Tensor) -> torch.Tensor:
    """
    In-place FWHT along the *last* dimension of `x`.
    Assumes size of last dim is a power of two.
    """
    x = x.detach().clone().contiguous()

    n = x.shape[-1]
    h = 1

    assert is_pow2(n)

    while h < n:
        x = x.view(*x.shape[:-1], n // (2 * h), 2, h)
        a = x[..., 0, :]
        b = x[..., 1, :]

        tmp = a.clone()
        a.copy_(tmp + b)
        b.copy_(tmp - b)

        x = x.view(*x.shape[:-3], n)
        h <<= 1

    return x



import functools
import os

import safetensors.torch
import torch

PKG_PATH = os.path.dirname(os.path.realpath(__file__))


@functools.cache
def load_optimal_grid_2_256(device="cpu", dtype=torch.float16):
    return safetensors.torch.load_file(
        os.path.join(PKG_PATH, "grids.safetensors"), device=device
    )["2_256"].to(dtype)



import torch


def higgs_dequantize_2_256_torch(x, grid):
    assert grid.dtype in (torch.float16, torch.bfloat16)
    assert x.dtype == torch.uint8
    assert grid.shape == (256, 2)

    (out_dim,) = x.shape
    assert out_dim > 0

    out = grid[x.to(torch.int32)]

    assert out.shape == (out_dim, 2)
    assert out.dtype == grid.dtype

    return out


def higgs_quantize_2_256_torch(x, grid, grid_norms):
    assert x.dtype == grid.dtype == grid_norms.dtype
    assert x.dtype in (torch.float16, torch.bfloat16)
    assert grid.shape == (256, 2), grid.shape
    assert grid_norms.shape == (256,), grid_norms.shape
    out_dim, in_dim = x.shape
    assert in_dim == 2
    assert out_dim > 0

    return torch.argmax(2 * x @ grid.T - grid_norms, dim=-1).to(torch.uint8)



import torch

from higgs_kernels.hadamard import hadamard_transform
from higgs_kernels.kernels import (
    higgs_dequantize_2_256_kernel,
    higgs_quantize_2_256_kernel,
)


def higgs_quantize_linear_2_256(weight, grid, group_size):
    assert weight.device == grid.device

    out_dim, in_dim = weight.shape
    assert in_dim % group_size == 0
    assert group_size % 2 == 0

    assert grid.shape == (256, 2)

    assert weight.dtype == grid.dtype
    assert weight.dtype in (torch.float16, torch.bfloat16)

    grid_norms = (torch.linalg.norm(grid, dim=-1) ** 2).to(grid.dtype)

    groups_per_input = in_dim // group_size

    weight = weight.reshape(out_dim, groups_per_input, group_size)
    scales = torch.linalg.norm(weight, dim=-1) / group_size**0.5
    assert scales.shape == (out_dim, groups_per_input)

    weight = weight / scales[:, :, None]

    weight = hadamard_transform(weight) / group_size**0.5

    weight = weight.reshape(-1, 2)

    quantized = higgs_quantize_2_256_kernel(weight, grid, grid_norms)
    assert quantized.shape == (out_dim * in_dim // 2,)
    assert quantized.dtype == torch.uint8

    return quantized, scales


def higgs_dequantize_linear_2_256(quantized, scales, grid, group_size):
    out_dim, groups_per_input = scales.shape
    in_dim = groups_per_input * group_size
    assert quantized.shape == (out_dim * in_dim // 2,)

    assert quantized.dtype == torch.uint8
    assert scales.dtype in (torch.float16, torch.bfloat16)
    assert grid.dtype == scales.dtype
    assert scales.device == quantized.device == grid.device

    output = higgs_dequantize_2_256_kernel(quantized, grid)
    assert output.shape == (out_dim * in_dim // 2, 2)

    output = output.reshape(out_dim, groups_per_input, group_size)

    output = hadamard_transform(output) / group_size**0.5

    output = output * scales[:, :, None]

    output = output.reshape(out_dim, in_dim)

    return output


def higgs_matmul_linear_2_256(input, quantized, scales, grid, group_size):
    dequantized = higgs_dequantize_linear_2_256(quantized, scales, grid, group_size)
    return torch.nn.functional.linear(input, dequantized)
